import os
import shutil
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict, Union, Optional

from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.serving.engines.base import EngineConfig, Command, EngineProcess, debug_msg

def set_full_permissions(path: str):
    """Recursively set full permissions for all files and directories in the given path."""
    for root, dirs, files in os.walk(path):
        for d in dirs:
            dir_path = os.path.join(root, d)
            os.chmod(dir_path, 0o777)  # Full permissions for directories
        for f in files:
            file_path = os.path.join(root, f)
            os.chmod(file_path, 0o777)  # Full permissions for files

def download_and_extract(version: str = "0.3.8", download_dir: str = ".", extract_dir: str = "ollama") -> Optional[str]:

    version = version.lstrip('v')
    url = f"https://github.com/ollama/ollama/releases/download/v{version}/ollama-linux-amd64.tgz"
    downloaded_file = os.path.join(download_dir, "ollama-linux-amd64.tgz")
    extract_path = os.path.join(download_dir, extract_dir)

    os.makedirs(download_dir, exist_ok=True)

    try:
        subprocess.run(["wget", url, "-O", downloaded_file], check=True)
        print(f"Downloaded {downloaded_file} from {url}")
    except subprocess.CalledProcessError as e:
        print(f"Error downloading file: {e}")
        return

    try:
        subprocess.run(["tar", "-xzvf", downloaded_file, "-C", download_dir], check=True)
        print(f"Extracted {downloaded_file} into {download_dir}")
    except subprocess.CalledProcessError as e:
        print(f"Error extracting file: {e}")
        return

    if os.path.exists(downloaded_file):
        os.remove(downloaded_file)
        print(f"Removed the downloaded file {downloaded_file}")

    if os.path.exists(extract_path):
        shutil.rmtree(extract_path)
        print(f"Removed the previous extracted directory {extract_path}")

    extracted_folder_name = "ollama-linux-amd64"
    extracted_folder_path = os.path.join(download_dir, extracted_folder_name)

    if os.path.exists(extracted_folder_path):
        os.rename(extracted_folder_path, extract_path)
        print(f"Renamed the extracted directory to {extract_path}")
    else:
        print(f"Expected directory {extracted_folder_path} not found. No renaming performed.")

    set_full_permissions(str(download_dir))

    return str((Path(download_dir) / "bin/ollama").absolute())


@dataclass(frozen=True, kw_only=True)
class OllamaEngineConfig(EngineConfig):
    ollama_version: str = field(default="0.3.8")
    model_download_dir_name: str = field(default="ollama-downloaded-models")
    model_artifact_key: str = field(default="model")
    root_folder_name: str = field(default="ollama")

    def setup_artifacts(self, local_dir: str = "/root/models") -> Dict[str, str]:
        # Example usage:
        download_dir = Path(local_dir) / self.root_folder_name
        ollama_cli = download_and_extract(self.ollama_version, str(download_dir))
        new_env = os.environ.copy()
        new_env.update({
          "OLLAMA_HOST": f"{self.host}:{self.port}",
          "OLLAMA_MODELS": str(download_dir / self.model_download_dir_name)
        })
        server = Command(name="ollama-serve", command=[ollama_cli, "serve"], env=new_env)
        server.start()
        download_model = Command(
            name="ollama-download-model",
            command=[ollama_cli, "pull", self.model],
            env=new_env,
            long_living=False
        )
        download_model.start()
        download_model.wait_and_log()
        server.stop()
        return {
            self.model_artifact_key: str(download_dir.absolute())
        }

    def to_run_command(self, context: PythonModelContext = None) -> Union[List[str], Command]:
        local_model_path = "/root/models"
        if context is not None:
            local_model_path = context.artifacts.get(self.model_artifact_key)
            set_full_permissions(str(local_model_path))
        ollama_root_dir = Path(local_model_path)
        bin_path = ollama_root_dir / "bin/ollama"
        new_env = os.environ.copy()
        new_env.update({
            "OLLAMA_HOST": f"{self.host}:{self.port}",
            "OLLAMA_MODELS": str(ollama_root_dir / self.model_download_dir_name)
        })
        return Command(name="ollama-serve", command=[bin_path, "serve"], env=new_env)

    def engine_pip_reqs(self, **kwargs) -> List[str]:
        return []


class OllamaEngineProcess(EngineProcess):

    @property
    def engine_name(self) -> str:
        return "ollama-engine"

    def health_check(self) -> bool:
        try:
            resp = self.server_http_client.get("/")
            return resp.status_code == 200
        except Exception as e:
            debug_msg(f"Health check failed with error {e}; server may not be up yet or crashed;")
            return False



import abc
import os
import random
import signal
import socket
import subprocess
import time
from dataclasses import field, dataclass
from typing import List, Optional, Dict, Union

import httpx
from filelock import FileLock
from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.version import get_mlflow_extensions_version


def debug_msg(msg: str):
    print(f"[DEBUG][pid:{os.getpid()}] {msg}")


def is_port_open(host: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        result = sock.connect_ex((host, port))
        return result == 0


@dataclass(kw_only=True)
class Command:
    name: str
    command: List[str]
    active_process: Optional[subprocess.Popen] = None
    long_living: bool = True
    env: Optional[Dict[str, str]] = None

    def start(self):
        if self.long_living is True:
            self.active_process = subprocess.Popen(self.command,
                                                   env=self.env or os.environ.copy(),
                                                   stdout=subprocess.PIPE,
                                                   stderr=subprocess.PIPE,
                                                   # ensure process is in another process group / session
                                                   preexec_fn=os.setsid, )
        else:
            self.active_process = subprocess.Popen(
                self.command,
                env=self.env or os.environ.copy(),
                preexec_fn=os.setsid,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,  # This will handle the output as text rather than bytes
                bufsize=1  # Line-buffered
            )

    def is_running(self) -> bool:
        if self.active_process is None:
            return False
        if self.active_process.returncode is not None:
            return False
        return True

    def wait_and_log(self):
        if self.active_process is None:
            print("Process has not been started.")
            return

        if self.long_living is True:
            raise ValueError("Unable to wait and log for long living process will hang the thread.")

        try:
            # Stream and print stdout
            for line in self.active_process.stdout:
                print(f"STDOUT: {line.strip()}")

            # Wait for the process to complete
            self.active_process.wait()

            # Check the exit code
            print('Exit Code:', self.active_process.returncode)

        except Exception as e:
            print(f"Error while waiting for logs: {e}")

    def stop(self):
        if self.is_running() is False:
            return
        # Send SIGTERM to the subprocess
        os.killpg(os.getpgid(self.active_process.pid), signal.SIGTERM)

        # Wait for a short while and check if the process has terminated
        time.sleep(5)
        if self.active_process.poll() is None:
            # If still running, send SIGKILL
            os.killpg(os.getpgid(self.active_process.pid), signal.SIGKILL)

        # # Optionally, wait for the process to terminate and get its exit status
        try:
            # Wait for the process to terminate, with a timeout
            stdout, stderr = self.active_process.communicate(timeout=10)
        except subprocess.TimeoutExpired:
            stdout, stderr = self.active_process.communicate()
            print("Timed out")

        try:
            print('STDOUT:', stdout.decode())
            print('STDERR:', stderr.decode())
        finally:
            print('Exit Code:', self.active_process.returncode)


@dataclass(frozen=True, kw_only=True)
class EngineConfig(abc.ABC):
    model: str
    host: str = field(default="0.0.0.0")
    port: int = field(default=9989)
    openai_api_path: int = field(default="v1")

    @abc.abstractmethod
    def to_run_command(self, context: PythonModelContext = None) -> Union[List[str], Command]:
        pass

    def default_pip_reqs(self, *,
                         filelock_version: str = "3.15.4",
                         httpx_version: str = "0.27.0",
                         mlflow_extensions_version: str = None,
                         **kwargs) -> List[str]:

        mlflow_extensions_version = mlflow_extensions_version or get_mlflow_extensions_version()
        if mlflow_extensions_version is None:
            mlflow_extensions = "mlflow-extensions"
        else:
            mlflow_extensions = f"mlflow-extensions=={mlflow_extensions_version}"
        return [f"httpx=={httpx_version}", *self.engine_pip_reqs(**kwargs),
                f"filelock=={filelock_version}", mlflow_extensions]

    @abc.abstractmethod
    def engine_pip_reqs(self, **kwargs) -> List[str]:
        pass

    @abc.abstractmethod
    def setup_artifacts(self, local_dir: str = "/root/models") -> Dict[str, str]:
        pass


class EngineProcess(abc.ABC):

    def __init__(self, *, config: EngineConfig):
        self._config = config
        self._lock = FileLock(f"{self.__class__.__name__}.txt.lock")
        self._server_http_client = httpx.Client(
            base_url=f"http://{self._config.host}:{self._config.port}",
            timeout=30
        )
        self._oai_http_client = httpx.Client(
            base_url=f"http://{self._config.host}:{self._config.port}/{self._config.openai_api_path}",
            timeout=300
        )
        self._proc = None

    @property
    def oai_http_client(self) -> httpx.Client:
        return self._oai_http_client

    @property
    def server_http_client(self) -> httpx.Client:
        return self._server_http_client

    @property
    @abc.abstractmethod
    def engine_name(self) -> str:
        pass

    @property
    def config(self) -> EngineConfig:
        return self._config

    @abc.abstractmethod
    def health_check(self) -> bool:
        pass

    def is_process_healthy(self) -> bool:
        if not is_port_open(self.config.host, self.config.port):
            return False

        return self.health_check()

    def stop_proc(self):
        with self._lock:
            if self.is_process_healthy() is True:
                subprocess.run(f"kill $(lsof -t -i:{self.config.port})", shell=True)

    # todo add local lora paths
    def start_proc(self, context: PythonModelContext = None):
        # kill process in port if already running
        time.sleep(random.randint(1, 5))
        debug_msg(f"Attempting to acquire Lock")
        with self._lock:
            debug_msg(f"Acquired Lock")
            if self.health_check() is False:
                proc_env = {"HOST": self.config.host, "PORT": str(self.config.host)}
                command = self.config.to_run_command(context)
                if isinstance(command, list):
                    self._proc = subprocess.Popen(self.config.to_run_command(context),
                                                  env=proc_env)
                elif isinstance(command, Command):
                    command.start()

                while self.is_process_healthy() is False:
                    debug_msg(f"Waiting for {self.engine_name} to start")
                    time.sleep(1)
            else:
                debug_msg(f"{self.engine_name} already running")



import json
from dataclasses import dataclass
from typing import List, Type, Optional, Iterator

import mlflow
from httpx import Response
from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.serving.engines.base import EngineProcess, debug_msg, EngineConfig


@dataclass
class CustomEngineServingResponse:
    status: int
    data: dict


class CustomServingEnginePyfuncWrapper(mlflow.pyfunc.PythonModel):

    # todo support lora modules
    def __init__(self,
                 *,
                 engine: Type[EngineProcess],
                 engine_config: EngineConfig,
                 endpoint="/chat/completions"):
        self._engine_klass: Type[EngineProcess] = engine
        self._engine_config: EngineConfig = engine_config
        self._engine: Optional[EngineProcess] = None
        # for convenience
        self._model_name = self._engine_config.model
        self._endpoint = endpoint
        self._artifacts = None

    @property
    def artifacts(self):
        if self._artifacts is None:
            raise ValueError("Artifacts not configured, run model.setup()")
        return self._artifacts

    def _request_model(self, req_str):
        response = self._engine.oai_http_client.post(self._endpoint, content=req_str)
        status_code = response.status_code
        return json.dumps({"status": status_code, "data": response.text})

    @staticmethod
    def iter_mlflow_predictions(response: Response) -> Iterator[CustomEngineServingResponse]:
        mlflow_response = response.json()
        predictions = mlflow_response.get("predictions", [])
        for prediction in predictions:
            prediction = json.loads(prediction)
            data = prediction.get("data", "")
            try:
                prediction_data = json.loads(data)
            except Exception as e:
                debug_msg(f"failed to parse data; got error: {str(e)}")
                prediction_data = data
            yield CustomEngineServingResponse(
                status=prediction.get("status"),
                data=prediction_data
            )

    def load_context(self, context: PythonModelContext):
        if self._engine is None:
            self._engine = self._engine_klass(config=self._engine_config)
        self._engine.start_proc(context)

    def predict(self, context, model_input: List[str], params=None) -> List[str]:
        return [self._request_model(req) for req in model_input]

    def _setup_artifacts(self, local_dir: str = "/root/models"):
        self._artifacts = self._engine_config.setup_artifacts(local_dir)
        return self._artifacts

    def get_pip_reqs(self, **kwargs):
        return self._engine_config.default_pip_reqs(**kwargs)

    def setup(self, *, local_dir="/root/models"):
        self._setup_artifacts(local_dir)
        debug_msg(f"Command to be run: {self._engine_config.to_run_command()}")



import json
import sys
from dataclasses import field, dataclass
from pathlib import Path
from typing import List, Dict, Optional, Union

from huggingface_hub import snapshot_download
from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.serving.engines.base import EngineConfig, debug_msg, EngineProcess, Command


@dataclass(frozen=True, kw_only=True)
class VLLMEngineConfig(EngineConfig):
    entrypoint_module: str = field(default="vllm.entrypoints.openai.api_server")
    enable_experimental_chunked_prefill: bool = field(default=False)
    max_num_batched_tokens: int = field(default=512)  # 512 is based on A100 ITL for llama model
    enable_prefix_caching: bool = field(default=False)
    vllm_command_flags: Dict[str, Optional[str]] = field(default_factory=dict)
    trust_remote_code: bool = field(default=False)
    max_model_len: Optional[int] = field(default=None)
    served_model_alias: Optional[str] = field(default=None)
    guided_decoding_backend: Optional[str] = field(default=None)
    model_artifact_key: str = field(default="model")
    verify_chat_template: bool = field(default=True)
    tokenizer_config_file: str = field(default="tokenizer_config.json")
    chat_template_key: str = field(default="chat_template")

    def _to_vllm_command(self, context: PythonModelContext = None) -> List[str]:
        local_model_path = None
        if context is not None:
            local_model_path = context.artifacts.get(self.model_artifact_key)
        flags = []
        skip_flags = ["--enable-chunked-prefill",
                      "--model",
                      "--max-num-batched-tokens",
                      "--enable-prefix-caching",
                      "--max-model-len",
                      "--trust-remote-code",
                      "--served-model-name",
                      "--guided-decoding-backend"]
        for k, v in self.vllm_command_flags.items():
            if k in skip_flags:
                debug_msg(f"Skipping flag {k} use the built in argument")
                continue
            flags.append(k)
            if v is not None:
                flags.append(v)
        if self.enable_experimental_chunked_prefill is True:
            flags.append("--enable-chunked-prefill")
            flags.append("--max-num-batched-tokens")
            flags.append(str(self.max_num_batched_tokens))
        if self.enable_prefix_caching is True:
            flags.append("--enable-prefix-caching")
        if self.trust_remote_code is True:
            flags.append("--trust-remote-code")
        if self.max_model_len is not None:
            flags.append("--max-model-len")
            flags.append(str(self.max_model_len))
        if self.guided_decoding_backend is not None:
            flags.append("--guided-decoding-backend")
            if self.guided_decoding_backend not in ["outlines", "lm-format-enforcer"]:
                raise ValueError(f"Invalid guided decoding backend {self.guided_decoding_backend}")
            flags.append(self.guided_decoding_backend)

        return [
            sys.executable,
            "-m",
            self.entrypoint_module,
            "--host",
            self.host,
            "--port",
            str(self.port),
            "--model",
            self.model if local_model_path is None else local_model_path,
            "--served-model-name",
            self.model if self.served_model_alias is None else self.served_model_alias,
            *flags,
        ]

    def to_run_command(self, context: PythonModelContext = None) -> Union[List[str], Command]:
        return self._to_vllm_command(context=context)

    def engine_pip_reqs(self, *,
                        vllm_version: str = "0.5.5",
                        lm_format_enforcer_version: str = "0.10.6",
                        outlines_version: str = "0.0.46") -> List[str]:
        default_installs = [f"vllm=={vllm_version}"]
        if self.guided_decoding_backend == "lm-format-enforcer":
            default_installs.append(f"lm-format-enforcer=={lm_format_enforcer_version}")
        if self.guided_decoding_backend == "outlines":
            default_installs.append(f"outlines=={outlines_version}")
        return default_installs

    def _hub_download_snapshot(self, repo_name: str, local_dir: str = "/root/models"):
        local_dir = local_dir.rstrip('/')
        model_local_path = f"{local_dir}/{repo_name}"
        snapshot_download(repo_id=self.model,
                          local_dir=model_local_path)
        return model_local_path

    def _setup_snapshot(self, local_dir: str = "/root/models"):
        return self._hub_download_snapshot(self.model, local_dir)

    def _setup_artifacts(self, local_dir: str = "/root/models"):
        local_path = self._setup_snapshot(local_dir)
        return {self.model_artifact_key: local_path}

    def _verify_chat_template(self, artifacts: Dict[str, str]):
        model_dir_path = Path(artifacts[self.model_artifact_key])
        tokenizer_config_file = model_dir_path / self.tokenizer_config_file
        assert tokenizer_config_file.exists(), f"Tokenizer config file not found at {str(tokenizer_config_file)}"
        with open(str(tokenizer_config_file), "r") as f:
            tokenizer_config = json.loads(f.read())
            chat_template = tokenizer_config.get(self.chat_template_key)
            if chat_template is None:
                raise ValueError(f"Chat template not found in tokenizer config file {str(tokenizer_config_file)}")

    def setup_artifacts(self, local_dir: str = "/root/models"):
        artifacts = self._setup_artifacts(local_dir)
        if self.verify_chat_template is True:
            self._verify_chat_template(artifacts)
        return artifacts


class VLLMEngineProcess(EngineProcess):

    @property
    def engine_name(self) -> str:
        return "vllm-engine"

    def health_check(self) -> bool:
        try:
            resp = self.server_http_client.get("/health")
            return resp.status_code == 200
        except Exception as e:
            debug_msg(f"Health check failed with error {e}; server may not be up yet or crashed;")
            return False
